#!/usr/bin/env python3

##################################################################################
# Pyprojectx                                                                     #
# https://github.com/houbie/pyprojectx                                           #
#                                                                                #
# Copyright (c) 2021 Ivo Houbrechts                                              #
#                                                                                #
# Licensed under the MIT license                                                 #
##################################################################################
import argparse
import os
import subprocess
import sys
from pathlib import Path
from venv import EnvBuilder

VERSION = "0.8.5"

PYPROJECTX_INSTALL_DIR_ENV_VAR = "PYPROJECTX_INSTALL_DIR"
PYPROJECTX_PACKAGE_ENV_VAR = "PYPROJECTX_PACKAGE"
PYPROJECT_TOML = "pyproject.toml"
DEFAULT_INSTALL_DIR = ".pyprojectx"

CYAN = "\033[96m"
BLUE = "\033[94m"
RED = "\033[91m"
RESET = "\033[0m"
if sys.platform.startswith("win"):
    os.system("color")


def run(args):
    try:
        options = get_options(args)
        pyprojectx_script = ensure_pyprojectx(options)
        explicit_options = []
        if not options.toml:
            explicit_options += ["--toml", str(options.toml_path)]
        if not options.install_dir:
            explicit_options += ["--install-dir", str(options.install_path)]

        subprocess.run([str(pyprojectx_script), *explicit_options, *args], check=True)
    except subprocess.CalledProcessError as e:
        raise SystemExit(e.returncode) from e


def get_options(args):
    options = arg_parser().parse_args(args)
    options.install_path = Path(
        options.install_dir
        or os.environ.get(PYPROJECTX_INSTALL_DIR_ENV_VAR, Path(__file__).with_name(DEFAULT_INSTALL_DIR))
    )
    options.toml_path = Path(options.toml) if options.toml else Path(__file__).with_name(PYPROJECT_TOML)
    if os.environ.get(PYPROJECTX_PACKAGE_ENV_VAR):
        options.version = "development"
        options.pyprojectx_package = os.environ.get(PYPROJECTX_PACKAGE_ENV_VAR)
    else:
        options.version = VERSION
        options.pyprojectx_package = f"pyprojectx~={VERSION}"
    options.verbosity = 0 if options.quiet or not options.verbosity else options.verbosity
    return options


def arg_parser():
    parser = argparse.ArgumentParser(
        description="Execute commands or aliases defined in the [tool.pyprojectx] section of pyproject.toml. "
        "Use the -i or --info option to see available tools and aliases.",
    )
    parser.add_argument("--version", action="version", version=VERSION)
    parser.add_argument(
        "--toml",
        "-t",
        action="store",
        help="The toml config file. Defaults to 'pyproject.toml' in the same directory as the pw script.",
    )
    parser.add_argument(
        "--install-dir",
        action="store",
        help=f"The directory where all tools (including pyprojectx) are installed; defaults to the"
        f"{PYPROJECTX_INSTALL_DIR_ENV_VAR} environment value if set, else '.pyprojectx'"
        f" in the same directory as the invoked pw script",
    )
    parser.add_argument(
        "--force-install",
        "-f",
        action="store_true",
        help="Force clean installation of the virtual environment used to run cmd, if any",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        dest="verbosity",
        help="Give more output. This option is additive and can be used up to 2 times.",
    )
    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Suppress output",
    )
    parser.add_argument(
        "--info",
        "-i",
        action="store_true",
        help="Show the configuration details of a command in stead of running it. "
        "If the command is not configured as tool or alias, a list with all available tools and aliases is shown.",
    )
    parser.add_argument(
        "--init",
        action="store_true",
        help="Create or prepare a pyproject.toml and pyprojectx wrapper scripts. "
        "Run with '--init help' to show available init options",
    )
    parser.add_argument(
        "command", nargs=argparse.REMAINDER, help="The command/alias with optional arguments to execute."
    )
    return parser


def ensure_pyprojectx(options):
    env_builder = EnvBuilder(with_pip=True)
    venv_dir = options.install_path.joinpath(
        "pyprojectx", f"{options.version}-py{sys.version_info.major}.{sys.version_info.minor}"
    )
    env_context = env_builder.ensure_directories(venv_dir)
    pyprojectx_script = Path(env_context.bin_path, "pyprojectx")
    pyprojectx_exe = Path(env_context.bin_path, "pyprojectx.exe")
    pip_cmd = [env_context.env_exe, "-m", "pip", "install"]

    if options.quiet:
        out = subprocess.DEVNULL
        pip_cmd.append("--quiet")
    else:
        out = sys.stderr

    if not pyprojectx_script.is_file() and not pyprojectx_exe.is_file():
        if not options.quiet:
            print(f"{CYAN}creating pyprojectx venv in {BLUE}{venv_dir}{RESET}", file=sys.stderr)
        env_builder.create(venv_dir)
        subprocess.run(
            pip_cmd + ["--upgrade", "pip"],
            stdout=out,
            check=True,
        )

        if not options.quiet:
            print(
                f"{CYAN}installing pyprojectx {BLUE}{options.version}: {options.pyprojectx_package} {RESET}",
                file=sys.stderr,
            )
        subprocess.run(pip_cmd + [options.pyprojectx_package], stdout=out, check=True)
    return pyprojectx_script


if __name__ == "__main__":
    run(sys.argv[1:])



import os
import re
import shutil
import subprocess
import sys
from pathlib import Path

import tomli

from pyprojectx.log import logger
from pyprojectx.wrapper.pw import BLUE, CYAN, DEFAULT_INSTALL_DIR, PYPROJECT_TOML, RESET

SCRIPT_EXTENSION = ".bat" if sys.platform.startswith("win") else ""
SCRIPT_PREFIX = "" if sys.platform.startswith("win") else "./"

HOME_DIR = Path(os.environ.get("PYPROJECTX_HOME_DIR", Path.home()))


def initialize(options):
    return INIT_COMMANDS.get(options.cmd, show_help)(options)


def show_help(_):
    """Show this help message."""
    print(f"{BLUE}Available --init commands:{RESET}", file=sys.stderr)
    for cmd, fn in INIT_COMMANDS.items():
        print(f"{CYAN}{cmd}{RESET}", fn.__doc__, file=sys.stderr)


def initialize_project(_):
    """Initialize pyprojectx setup in the current working directory.
    If pyproject.toml already exists and doesn't yet contain a tool.pyprojectx section,
    an example section will be appended.
    """
    _initialize_template("project-template.toml")
    _print_usage()


def initialize_poetry(options):
    """Initialize a poetry project in the current working directory, along with pyprojectx scripts."""
    logger.info("copying poetry.toml")
    shutil.copy2(Path(__file__).with_name("poetry.toml"), ".")
    _initialize_build_tool("poetry", options)


def initialize_pdm(options):
    """Initialize a PDM project in the current working directory, along with pyprojectx scripts."""
    _initialize_build_tool("pdm", options)


def _initialize_build_tool(tool, options):
    template = f"{tool}-template.toml"
    _initialize_template(template, toml_file=template)

    logger.info("installing %s...", tool)
    proc = subprocess.run(
        f"{SCRIPT_PREFIX}pw --toml {template} {tool} --version", shell=True, check=True, capture_output=True
    )
    version = re.search(r"(\d+\.)+(\d+)", proc.stdout.decode("utf-8"))[0]
    old_requirement = f"{tool}>=1.1"
    new_requirement = f"{tool}=={version}"
    logger.info("setting version in %s : %s", PYPROJECT_TOML, new_requirement)
    _replace_in_file(old_requirement, new_requirement, template)
    subprocess.run(
        f"{SCRIPT_PREFIX}pw --toml {template} {tool} init " + " ".join(options.cmd_args), shell=True, check=True
    )
    logger.debug("appending template to %s...", PYPROJECT_TOML)
    with open(template, "rt") as src:
        with open(PYPROJECT_TOML, "at") as dest:
            dest.write(src.read())
    _print_usage()
    os.remove(template)
    print(
        f"\n{BLUE}You can run all {CYAN}{tool}{BLUE} commands by typing {RESET}{SCRIPT_PREFIX}pw {BLUE} in front",
        file=sys.stderr,
    )
    print(f"Example: {RESET}{SCRIPT_PREFIX}pw {tool} update", file=sys.stderr)
    print(f"{BLUE}Or use the shorter aliases like {RESET}{SCRIPT_PREFIX}pw install {BLUE}and", file=sys.stderr)
    print(
        f"{RESET}{SCRIPT_PREFIX}pw run {BLUE}to install your project or run a script with {CYAN}{tool}{RESET}",
        file=sys.stderr,
    )


def _initialize_template(template_name, toml_file=PYPROJECT_TOML):
    wrapper_dir = Path(__file__).parent.parent.joinpath("wrapper")
    target_pw = Path("pw")
    if not target_pw.exists():
        logger.info("copying wrapper scripts")
        shutil.copy2(wrapper_dir.joinpath("pw.py"), target_pw)
        shutil.copy2(wrapper_dir.joinpath("pw.bat"), ".")
    else:
        logger.info("wrapper scripts already present")
    target_toml = Path(toml_file)
    template = Path(__file__).with_name(template_name)
    if not target_toml.exists():
        logger.info("copying %s template", template_name)
        shutil.copy2(template, target_toml)
    else:
        with open(target_toml, "a+") as dst:
            toml_dict = tomli.load(dst)
            if not toml_dict.get("tool", {}).get("pyprojectx"):
                with open(template, "rt") as src:
                    logger.info("appending template to %s", toml_file)
                    dst.write(src.read())


def _replace_in_file(old, new, file):
    with open(file, "rt") as f:
        text = f.read().replace(old, new)
    with open(file, "wt") as f:
        f.write(text)


def _print_usage():
    print(f"{BLUE}Pyprojectx scripts are installed in the current directory.", file=sys.stderr)
    print("You can add pw and pw.bat under version control if applicable.", file=sys.stderr)
    if sys.platform.startswith("win"):
        print(f"When using git, run {RESET}git add pw pw.bat && git update-index --chmod=+x pw'", file=sys.stderr)
    print(
        f"{BLUE}Run {RESET}{SCRIPT_PREFIX}pw --info -{BLUE}"
        f" to see the available tools and aliases in your project.{RESET}",
        file=sys.stderr,
    )


def initialize_global(options):
    """Initialize the global pyprojectx setup in your home directory.
    Use '--init global --force' to overwrite.
    """
    global_dir = HOME_DIR.joinpath(DEFAULT_INSTALL_DIR, "global")
    wrapper_dir = Path(__file__).parent.parent.joinpath("wrapper")
    logger.debug("creating global directory %s", global_dir)
    os.makedirs(global_dir, exist_ok=True)

    target_pw = global_dir.joinpath("pw")
    if target_pw.exists() and "--force" not in options.cmd_args:
        print(f"{target_pw} {BLUE} already exists, use '--init global --force' to overwrite{RESET}", file=sys.stderr)
        return

    shutil.copy2(wrapper_dir.joinpath("pw.py"), target_pw)
    shutil.copy2(wrapper_dir.joinpath(f"px{SCRIPT_EXTENSION}"), global_dir.parent)
    shutil.copy2(wrapper_dir.joinpath(f"pxg{SCRIPT_EXTENSION}"), global_dir.parent)
    target_toml = global_dir.joinpath(PYPROJECT_TOML)
    if not target_toml.exists():
        shutil.copy2(Path(__file__).with_name("global-template.toml"), target_toml)

    print(f"{BLUE}Pyprojectx scripts are installed in your home directory.", file=sys.stderr)
    if sys.platform.startswith("win"):
        print(
            f"Add the scripts to your path, by adding\n"
            f"{RESET}{global_dir.parent.absolute()}\n"
            f"{BLUE}to the path environment variable.{RESET}",
            file=sys.stderr,
        )
    else:
        print(
            f"Add the scripts to your PATH, f.e. by appending following to your shell's profile"
            f" ({RESET}~/.profile{BLUE}, {RESET}~/.zshrc{BLUE}, {RESET}~/..bashrc{BLUE}, ...):\n"
            f"{RESET}export PATH=$PATH:{global_dir.parent.absolute()}, file=sys.stderr"
        )
        print(
            f"{BLUE}Run {RESET}px --info -{BLUE} to see the available tools and aliases in your project.",
            file=sys.stderr,
        )
        print(
            f"Run {RESET}pxg --info -{BLUE} to see the available tools and aliases in the global pyprojectx.{RESET}",
            file=sys.stderr,
        )


INIT_COMMANDS = {
    "help": show_help,
    "project": initialize_project,
    "poetry": initialize_poetry,
    "pdm": initialize_pdm,
    "global": initialize_global,
}



import subprocess
import sys
from typing import Iterable, List, Optional, Union

from pyprojectx.config import Config
from pyprojectx.env import IsolatedVirtualEnv
from pyprojectx.initializer.initializers import initialize
from pyprojectx.log import logger, set_verbosity
from pyprojectx.wrapper import pw


def main() -> None:
    _run(sys.argv)


def _run(argv: List[str]) -> None:
    options = _get_options(argv[1:])
    if options.init:
        initialize(options)
        return

    config = Config(options.toml_path)
    cmd = options.cmd
    if options.info:
        config.show_info(cmd)
        return

    if not cmd:
        pw.arg_parser().print_help(file=sys.stderr)
        raise SystemExit(1)

    aliases = config.find_aliases(cmd)
    logger.debug("Matching aliases for %s: %s", cmd, ", ".join(aliases))
    if aliases:
        if len(aliases) > 1:
            print(
                f"{pw.RED}'{cmd}' is ambiguous",
                file=sys.stderr,
            )
            print(
                f"{pw.BLUE}Candidates are:{pw.RESET}",
                file=sys.stderr,
            )
            print(", ".join(aliases), file=sys.stderr)
            raise SystemExit(1)
        tool, alias_cmd = config.get_alias(aliases[0])
        if alias_cmd:
            _run_alias(
                tool,
                alias_cmd,
                argv,
                requirements=config.get_tool_requirements(tool),
                options=options,
            )
    elif config.is_tool(cmd):
        _run_in_tool_venv(
            cmd,
            [cmd, *options.cmd_args],
            requirements=config.get_tool_requirements(cmd),
            options=options,
        )
    else:
        config.show_info(cmd, error=True)
        raise SystemExit(1)


def _run_alias(
    tool: Optional[str],
    alias_cmd: str,
    argv: List[str],
    requirements: Iterable[str],
    options,
) -> None:
    logger.debug("Running alias command, tool: %s, command: %s, arguments: %s", tool, alias_cmd, options.cmd_args)
    full_cmd = " ".join([_resolve_alias_references(alias_cmd, options.cmd, argv)] + options.cmd_args)
    if tool:
        _run_in_tool_venv(
            tool,
            full_cmd,
            requirements=requirements,
            options=options,
        )
    else:
        try:
            subprocess.run(full_cmd, shell=True, check=True)
        except subprocess.CalledProcessError as e:
            raise SystemExit(e.returncode) from e


def _run_in_tool_venv(
    tool: str,
    full_cmd: Union[str, List[str]],
    requirements: Iterable[str],
    options,
) -> None:
    logger.debug("Running tool command in virtual environment, tool: %s, full command: %s", tool, full_cmd)
    venv = IsolatedVirtualEnv(options.venvs_dir, tool, requirements)
    if not venv.is_installed or options.force_install:
        try:
            venv.install(quiet=options.quiet)
        except subprocess.CalledProcessError as e:
            print(
                f"{pw.RED}PYPROJECTX ERROR: installation of '{tool}' failed with exit code {e.returncode}{pw.RESET}",
                file=sys.stderr,
            )
            raise SystemExit(e.returncode) from e

    try:
        venv.run(full_cmd)
    except subprocess.CalledProcessError as e:
        raise SystemExit(e.returncode) from e


def _resolve_alias_references(alias_cmd: str, cmd: str, argv: List[str]) -> str:
    cmd_index = argv.index(cmd)
    replacement = " ".join(argv[:cmd_index]) + " "
    return alias_cmd.replace("pw@", replacement)


def _get_options(args):
    options = pw.get_options(args)
    if options.command:
        options.cmd, *options.cmd_args = options.command
    else:
        options.cmd = None
        options.cmd_args = []
    options.venvs_dir = options.install_path.joinpath("venvs")
    set_verbosity(options.verbosity)
    logger.debug("Parsed cli arguments: %s", options)
    return options



import logging

logger = logging.getLogger(__name__)


def set_verbosity(verbosity: int):
    if verbosity == 1:
        logging.getLogger().setLevel(logging.INFO)
    if verbosity > 1:
        logging.getLogger().setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)
    logging.basicConfig()



"""
Creates and manages isolated build environments.
"""
import hashlib
import os
import shutil
import subprocess
import sys
import tempfile
from collections import OrderedDict
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union

import virtualenv

from pyprojectx.log import logger


def calculate_path(base_path: Path, name: str, requirements: Iterable[str]) -> Path:
    md5 = hashlib.md5()
    for req in requirements:
        md5.update(req.strip().encode())
    return Path(
        base_path, f"{name.lower()}-{md5.hexdigest()}-py{sys.version_info.major}.{sys.version_info.minor}"
    ).absolute()


class IsolatedVirtualEnv:
    """
    Encapsulates the location and installation of an isolated virtual environment.
    """

    def __init__(self, base_path: Path, name: str, requirements: Iterable[str]) -> None:
        """
        :param base_path: The base path for all environments
        :param name: The name for the environment
        :param requirements: The requirements to install in the environment
        """
        self._name = name
        self._base_path = base_path
        self._requirements = requirements
        self._path = calculate_path(base_path, name, self._requirements)
        self._scripts_path_file = self._path.joinpath(".scripts_path")
        self._executable = None

    @property
    def name(self) -> str:
        """
        The name of the isolated environment.
        """
        return self._name

    @property
    def path(self) -> Path:
        """
        The location of the isolated environment.
        """
        return self._path

    @property
    def executable(self) -> Optional[Path]:
        """
        The location of the Python executable of the isolated environment.
        """
        return self._executable

    @property
    def scripts_path(self) -> Optional[Path]:
        """
        The location of the scripts directory.
        """
        if self._scripts_path_file.exists():
            with open(self._scripts_path_file, "r") as sf:
                return Path(sf.readline())
        return None

    @property
    def is_installed(self) -> bool:
        return self.scripts_path and self.scripts_path.is_dir()

    def install(self, quiet=False) -> None:
        """
        Create the virtual environment and install requirements
        """
        logger.debug("Installing IsolatedVirtualEnv in %s", self.path)
        scripts_dir = self._create_virtual_env()
        self._install_requirements(quiet)
        with open(self._scripts_path_file, "w") as sf:
            sf.write(str(scripts_dir))

    def _create_virtual_env(self) -> Path:
        cmd = [str(self.path), "--no-setuptools", "--no-wheel", "--activators", ""]
        logger.debug("Calling virtualenv.cli_run: %s", " ".join(cmd))
        result = virtualenv.cli_run(cmd, setup_logging=False)
        scripts_dir = result.creator.script_dir
        self._executable = result.creator.exe
        return scripts_dir

    def _install_requirements(self, quiet=False):
        logger.info("Installing packages in isolated environment... (%s)", ", ".join(sorted(self._requirements)))
        # pip does not honour environment markers in command line arguments
        # but it does for requirements from a file
        with tempfile.NamedTemporaryFile("w+", prefix="build-reqs-", suffix=".txt", delete=False) as req_file:
            req_file.write(os.linesep.join(self._requirements))
        try:
            cmd = [
                str(self._executable),
                "-Im",
                "pip",
                "install",
            ]
            if quiet:
                cmd.append("--quiet")
            cmd += [
                "--use-pep517",
                "--no-warn-script-location",
                "-r",
                os.path.abspath(req_file.name),
            ]
            subprocess.run(
                cmd,
                stdout=sys.stderr,
                check=True,
            )
        finally:
            os.remove(req_file.name)

    def remove(self):
        """
        Remove the entire virtual environment
        """
        logger.info("Removing isolated environment in %s", self.path)
        shutil.rmtree(self.path)

    def run(self, cmd: Union[str, List[str]]) -> subprocess.CompletedProcess:
        """
        Run a command inside the virual environment.
        :param cmd: The command string to run
        :return: The subprocess.CompletedProcess instance
        """
        logger.info("Running command in isolated venv %s: %s", self.name, cmd)
        if isinstance(cmd, List):
            extension = ".exe" if sys.platform.startswith("win") else ""
            cmd[0] = str(self.scripts_path.joinpath(cmd[0] + extension))
            shell = False
        else:
            shell = True

        paths: Dict[str, None] = OrderedDict()
        paths[str(self.scripts_path)] = None
        if "PATH" in os.environ:
            paths.update((i, None) for i in os.environ["PATH"].split(os.pathsep))
        extra_environ = {"PATH": os.pathsep.join(paths)}
        env = os.environ.copy()
        env.update(extra_environ)
        logger.debug("Environment for running command: %s", env)
        return subprocess.run(cmd, env=env, shell=shell, check=True)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyprojectx', 'pyprojectx.initializer', 'pyprojectx.wrapper']

package_data = \
{'': ['*']}

install_requires = \
['tomli>=1.2.2,<2.0.0', 'virtualenv>=20.10.0,<21.0.0']

entry_points = \
{'console_scripts': ['pyprojectx = pyprojectx.cli:main']}

setup_kwargs = {
    'name': 'pyprojectx',
    'version': '0.8.5',
    'description': 'Execute scripts from pyproject.toml, installing tools on-the-fly',
    'long_description': '# pyprojectx\n\nExecute scripts from pyproject.toml, installing tools on-the-fly\n\nGetting started with a Python project should be a one-liner:\n```shell\ngit clone https://github.com/houbie/pyprojectx.git && cd pyprojectx && ./pw build\n```\n\n![Cast](https://raw.githubusercontent.com/houbie/pyprojectx/main/docs/poetry-build-cast.svg)\n\nPyprojectx provides a CLI wrapper for automatic installation of Python tools:\n* Make it be a breeze for others to get started with your project or tutorial\n* Get reproducible builds by always using the correct versions of your build tools\n* Plays well with build tools like [Poetry](https://python-poetry.org/) and [PDM](https://pdm.fming.dev/)\n\nPyprojectx brings `npm run` to Python with:\n* less keystrokes\n* isolated dependencies: tools are not required to be development dependencies\n\n## Installation\nCopy the [wrapper scripts](https://github.com/houbie/pyprojectx/releases/latest/download/wrappers.zip)\ninto the root of your project.\n\nPython >= 3.7 must be available on your PATH.\n\n* osx / linux :\n```shell\ncurl -LO https://github.com/houbie/pyprojectx/releases/latest/download/wrappers.zip && unzip wrappers.zip && rm -f wrappers.zip\n```\n\n* windows: unpack the [wrappers zip](https://github.com/houbie/pyprojectx/releases/latest/download/wrappers.zip)\n\n**NOTE** On windows you need to explicitly mark the osx/linux script as executable before adding it to version control.\nWhen using git:\n```shell\ngit add pw pw.bat\ngit update-index --chmod=+x pw\n```\n\n## Configuration\nAdd the _tool.pyprojectx_ section inside _pyproject.toml_ in your project\'s root directory.\n\nEach entry has the form `tool = "pip-requirements"`, where _pip-requirements_ adheres to the\n[requirements file format](https://pip.pypa.io/en/stable/reference/requirements-file-format/).\n\nExample:\n```toml\n[tool.pyprojectx]\n# require a specific poetry version\npoetry = "poetry==1.1.11"\n# use the latest black\nblack = "black"\n# install flake8 in combination with plugins; one line per requirement \nflake8 = """\nflake8\nflake8-bandit\npep8-naming\nflake8-isort\nflake8-pytest-style"""\n```\n\nThe _tool.pyprojectx.aliases_ section can contain optional commandline aliases in the form\n\n`alias = [@tool_key:] command`\n\nExample:\n```toml\n[tool.pyprojectx.alias]\n# convenience shortcuts\nrun = "poetry run"\ntest = "poetry run pytest"\npylint = "poetry run pylint"\n\n# tell pw that the bandit binary is installed as part of flake8\nbandit = "@flake8: bandit my_package tests -r"\n\n# simple shell commands (watch out for variable substitutions and string literals containing whitespace or special characters )\nclean = "rm -f .coverage && rm -rf .pytest_cache"\n\n# when running an alias from within another alias, prefix it with `pw@`\ncheck = "pw@pylint && pw@test"\n```\n\nAliases can be invoked as is or with extra arguments:\n```shell\n./pw bandit\n\n./pw poetry run my-script\n# same as above, but using the run alias\n./pw run my-script\n```\n\n## Isolation\nEach tool gets installed in an isolated virtual environment.\n\nThese are all located in the _.pyprojectx_ directory of your project\n(where _pyproject.toml_ is located).\n\n# Usage\nInstead of calling the commandline of a tool directly, prefix it with `path\\to\\pw`.\n\nExamples:\n```shell\n./pw poetry add -D pytest\ncd src\n../pw black *.py\n```\n\n... or on Windows:\n```shell\npw poetry add -D pytest\ncd src\n..\\pw black *.py\n```\n\nWhen you run an alias on the command-line, you don’t have to provide the full name of the alias.\nYou only need to provide enough of the alias name to uniquely identify it.\nFor example, it’s likely `./pw che` is enough to run the _check_ alias if it exists.\n\nYou can use camel case patterns for more complex abbreviations. These patterns are expanded to match camel case\nand kebab case names. For example the pattern foBa (or even fB) matches fooBar and foo-bar.\n\n**NOTE** aliases can hide tool commands. F.e. a _pylint-src_ alias will hide the _pylint_ tool and will always run\nwhen running `./pw pylint some args`. This can be solved by making sure that no alias name starts with _pylint_ or by\nadding _pylint_ as an alias: `pylint="pylint"`.\n\nCheck _pw_ specific options with `pw --help`\n\n## Project initialization\nInitialize a new or existing project with the _--init_ option.\n* `./pw --init project`: add pyprojectx example sections to an existing or new _pyproject.toml_ in the current directory.\n* `./pw --init poetry`: initialize a [Poetry](https://python-poetry.org/) project and add pyprojectx example sections to _pyproject.toml_.\n* `./pw --init pdm`: initialize a [PDM](https://pdm.fming.dev/) project and add pyprojectx example sections to _pyproject.toml_.\n* `./pw --init global`: initialize a global pyprojectx setup in your home directory that can be accessed with the _pgx_ command.\nThis turns pyprojectx into a lightweight [Pipex](https://pypa.github.io/pipx/) alternative.\n* `./pw --init` or `./pw --init help`:  show initializer help.\n\n## px: run pw from any subdirectory\nIf you want to avoid typing `./pw` (or `../pw` when in a subdirectory), you can install the _px_\nscript in your home directory with `./pw --init global` (or `pw --init global` on Windows) and\nadd _~/.pyprojectx_ to your PATH.\n\nFrom then on, you can replace _pw_ with _px_ and invoke it from any (sub)directory containing the _pw_ script.\n```shell\ncd my-pyprojectx-project\npx test\ncd tests\npx test sometest.py\n```\n\n## Uninstall / cleaning up\n* Delete the _.pyprojectx_ directory in your project\'s root.\n* Delete the global _.pyprojectx_ directory in your home directory.\n\n## Why yet another tool when we already have pipx etc.?\n* As Python noob I had hard times setting up a project and building existing projects\n* There is always someone in the team having issues with his setup, either with a specific tool, with Homebrew, pipx, ...\n* Adding tools as dev dependencies impacts your production dependencies and can even lead to unresolvable conflicts\n* Different projects often require different versions of the same tool\n\n## Best practices\n* Separate your tools from your project dependencies\n* Use a build tool with decent dependency management that locks all dependencies,\n  f.e. [Poetry](https://python-poetry.org/) or [PDM](https://pdm.fming.dev/)\n* Pin down the version of your build tool to prevent the "project doesn\'t build anymore" syndrome.\n  Eventually a new version of the build tool with breaking changes will be released.\n* There is a category of tools that you don\'t want to version: tools that interact with changing environments.\n\n## Examples\n* This project (using Poetry)\n* [px-demo](https://github.com/houbie/px-demo)\n\n## Development\n* Build/test:\n```shell\ngit clone git@github.com:houbie/pyprojectx.git\ncd pyprojectx\n./pw build\n```\n\n* Set the path to pyprojectx in the _PYPROJECTX_PACKAGE_ environment variable\n to use your local pyprojectx copy in another project.\n```shell\n# *nix\nexport PYPROJECTX_PACKAGE=path/to/pyprojectx\n# windows\nset PYPROJECTX_PACKAGE=path/to/pyprojectx\n```\n',
    'author': 'Houbie',
    'author_email': 'ivo@houbrechts-it.be',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/houbie/pyprojectx',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import re
import sys
from itertools import zip_longest
from pathlib import Path
from typing import Iterable, List, Optional, Tuple

import tomli

from pyprojectx.wrapper.pw import BLUE, CYAN, RESET


class Config:
    """
    Encapsulates the PyprojectX config inside a toml file
    """

    def __init__(self, toml_path: Path) -> None:
        """
        :param toml_path: The toml config file
        """
        self._toml_path = toml_path
        try:
            with open(self._toml_path, "rb") as f:
                toml_dict = tomli.load(f)
                self._tools = toml_dict.get("tool", {}).get("pyprojectx", {})
                self._aliases = self._tools.get("aliases", {})
                self._os_aliases = self._merge_os_aliases()
        except Exception as e:
            raise Warning(f"Could not parse {self._toml_path}: {e}") from e

    def show_info(self, cmd, error=False):
        tool, alias = self.get_alias(cmd)
        out = sys.stderr if error else sys.stdout
        if alias:
            print(f"{cmd}{BLUE} is an alias in {CYAN}{self._toml_path.absolute()}", file=sys.stderr)
            if tool:
                print(f"{BLUE}and runs in the {CYAN}{tool}{BLUE} tool context", file=sys.stderr)
            print(f"{BLUE}command:{RESET}", file=sys.stderr)
            print(alias, file=out)
        elif self.is_tool(cmd):
            print(f"{cmd}{BLUE} is a tool in {CYAN}{self._toml_path.absolute()}", file=sys.stderr)
            print(f"{BLUE}requirements:{RESET}", file=sys.stderr)
            print(self.get_tool_requirements(cmd), file=out)
        else:
            if cmd:
                print(
                    f"{cmd}{BLUE} is not configured as tool or alias in {self._toml_path.absolute()}:{RESET}",
                    file=sys.stderr,
                )
            print(f"{BLUE}available aliases:{RESET}", file=sys.stderr)
            print("\n".join(self._aliases.keys()), file=out)
            print(f"{BLUE}available tools:{RESET}", file=sys.stderr)
            print("\n".join(self._tools.keys() - ["aliases", "os"]), file=out)

    def get_tool_requirements(self, key) -> Iterable[str]:
        """
        Get the requirements (dependencies) for a configured tool in the [tool.pyprojectx] section.
        The requirements can either be specified as a string or a list of strings.
        A multiline string is treated as one requirement per line.
        :param key: The key (tool name) to look for
        :return: the requirements as a list or None if key is not found
        """
        requirements = self._tools.get(key)
        if isinstance(requirements, list):
            return requirements
        if isinstance(requirements, str):
            return requirements.splitlines()
        return []

    def is_tool(self, key) -> bool:
        """
        Check whether a key (tool name) exists in the [tool.pyprojectx] section.
        :param key: The key (tool name) to look for
        :return: True if the key exists in the [tool.pyprojectx] section.
        """
        return self._tools.get(key) is not None

    def get_alias(self, key) -> Tuple[Optional[str], Optional[str]]:
        """
        Get an alias command configured in the [tool.pyprojectx.alias] section.
        The alias is considered to be part of a tool if its command starts with the name of the tool
        or if the the command starts with '@tool-name:'.
        :param key: The key (name) of the alias
        :return: A tuple containing the corresponding tool (or None) and
         the alias command (without the optional @tool-name part), or None if there is no alias with the given key.
        """
        alias_cmd = self._aliases.get(key)
        if not alias_cmd:
            return None, None
        if re.match(r"^@?[\w|-]+\s*:\s*", alias_cmd):
            tool, alias_cmd = re.split(r"\s*:\s*", alias_cmd, maxsplit=1)
            if tool.startswith("@"):
                tool = tool[1:]
            if not self.is_tool(tool):
                raise Warning(f"Invalid alias {key}: '{tool}' is not defined in [tool.pyprojectx]")
            return tool, alias_cmd
        tool = alias_cmd.split()[0]
        if self.is_tool(tool):
            return tool, alias_cmd
        return None, alias_cmd

    def find_aliases(self, abbrev: str) -> List[str]:
        """
        Find all alias keys that match the abbreviation.
        The abbreviation can use camel case patterns that are expanded to match camel case and kebab case names.
        For example the pattern foBa (or even fB) matches fooBar and foo-bar.
        If the abbreviation is exactly equal to an alias key, only that key is returned, even if the abbreviation
        matches other alias keys.
        :param abbrev: abbreviated or full alias key to search for
        :return: a list of matching alias keys
        """
        if abbrev in self._aliases.keys():
            return [abbrev]

        return [key for key in self._aliases.keys() if camel_match(abbrev, key)]

    def __repr__(self):
        return str(self._tools)

    def _merge_os_aliases(self):
        aliases = self._tools.get("aliases", {})
        os_dict = self._tools.get("os", {})
        for os_key in os_dict:
            if sys.platform.startswith(os_key) and isinstance(os_dict[os_key], dict):
                aliases.update(os_dict[os_key].get("aliases", {}))
        return aliases


def camel_match(abbrev, key):
    abbrev_parts = to_camel_parts(abbrev)
    full_parts = to_camel_parts(key)
    return all(f.startswith(s) for s, f in zip_longest(abbrev_parts, full_parts, fillvalue=""))


def to_camel_parts(key):
    if not key:
        return []
    if len(key) < 2:
        return [key]
    camel = re.sub(r"(-\w)", lambda m: m.group(0)[1].upper(), key)
    return filter(len, re.split("([A-Z][^A-Z]*)", camel[0].lower() + camel[1:]))
